import argparse
import os
import sys
import traceback

from hombre.constants import PrecheckPlugin
from hombre.nsx_config import full_sync, cleanup
from hombre.postinstall import register
from hombre.precheck.plugin_base import PluginBase
from hombre.utils import upgrade_pvc_size


# initialize plugins for precheck
PRECHECK_PLUGINS = {p.id: p for p in PluginBase.plugins}

_COMMAND_RESIZE_PVC = "resize-pvc"


def main():
    parser = argparse.ArgumentParser(description='Python utilities module for NSX Intelligence')
    parser.add_argument('--register-ui-plugin', action="store_true", required=False,
                        help="Invoke API on NSX manager to register Intelligence UI plugin")
    parser.add_argument('--unregister-ui-plugin', action="store_true", required=False,
                        help="Invoke API on NSX manager to UN-register Intelligence UI plugin")
    parser.add_argument('--enable-intelligence', action="store_true", required=False,
                        help="Invoke API on NSX manager to enable Intelligence")
    parser.add_argument('--disable-intelligence', action="store_true", required=False,
                        help="Invoke API on NSX manager to deactivate Intelligence")
    parser.add_argument('--validate-transport-node-count', action="store_true", required=False,
                        help="Invoke API on NSX manager to retrieve count of transport nodes "
                             "and validate that it is within limits")
    parser.add_argument('--validate-license', action="store_true", required=False,
                        help="Invoke API on NSX manager to retrieve licenses and "
                             "validate if minimum requirements are met")
    parser.add_argument('--validate-platform-type', action="store_true", required=False,
                        help="Invoke API on monitor app to validate if minimum platform "
                             "form factor requirement is satisfied")
    parser.add_argument('--validate-advanced-license', action="store_true", required=False,
                        help="Invoke API on NSX manager to retrieve licenses and "
                             "check if advanced license is installed")
    parser.add_argument('--validate-capacity', action="store_true", required=False,
                        help="Invoke feature capacity check API on 'cluster-api' to validate there exists enough"
                             " buffer capacity post feature enablement.")
    parser.add_argument('--register-site', action="store_true", required=False,
                        help="Invoke API(s) on NSX Manager to register site with NSX Intelligence")
    parser.add_argument('--full-sync', action="store", required=False,
                        help="Invoke API to start full sync via nsx config")
    parser.add_argument('--del-config-data', action="store_true", required=False,
                        help="Delete config data from postgres tables")
    parser.add_argument('--kindtest', action="store_true", required=False,
                        help='Register a virtual site with NSX Intelligence during kindtest')
    parser.add_argument('--site-id', action='store', required=False,
                        help='Virtual site ID to register with NSX Intelligence')

    subparsers = parser.add_subparsers(help="Additional commands", dest="command")

    # Specific options for resize-pvc
    parser_resize_pvc = subparsers.add_parser(_COMMAND_RESIZE_PVC)
    parser_resize_pvc.add_argument(
        "--statefulset",
        "-sts",
        type=str,
        required=True,
        help="The name of stateful set to be upgraded",
    )
    parser_resize_pvc.add_argument(
        "--label", "-l", type=str, required=True, help="The label for the stateful set and pvc"
    )
    parser_resize_pvc.add_argument(
        "--namespace", "-ns", type=str, default="nsxi-platform", help="Name space of the sts"
    )
    parser_resize_pvc.add_argument("--size", "-s", type=str, required=True, help="New pvc size")

    args = parser.parse_args()

    if in_kindtest():
        if (args.register_ui_plugin or
                args.unregister_ui_plugin or
                args.enable_intelligence or
                args.disable_intelligence):
            print('in kindtest mode, not registering UI plugins or enabling/disabling intelligence')
            return

    if args.register_ui_plugin:
        register.ui_plugin()
    elif args.unregister_ui_plugin:
        register.ui_plugin(unregister=True)
    elif args.enable_intelligence:
        register.cluster()
    elif args.disable_intelligence:
        register.cluster(unregister=True)
    elif args.register_site:
        register.site()
    # TODO(tandonp): use kindtest env variables
    elif args.kindtest:
        if not args.site_id:
            raise Exception("site-id must be provided in kindtest mode")
        register.site_kindtest(args.site_id)
    elif args.validate_transport_node_count:
        tn_plugin = PRECHECK_PLUGINS[PrecheckPlugin.TRANSPORT_NODES]()
        tn_plugin.check()
    elif args.validate_license:
        license_plugin = PRECHECK_PLUGINS[PrecheckPlugin.LICENSE]()
        license_plugin.check()
    elif args.validate_platform_type:
        platform_type_plugin = PRECHECK_PLUGINS[PrecheckPlugin.PLATFORM_TYPE]()
        platform_type_plugin.check()
    elif args.validate_advanced_license:
        advanced_license_plugin = PRECHECK_PLUGINS[PrecheckPlugin.ADVANCED_LICENSE]()
        advanced_license_plugin.check()
    elif args.validate_capacity:
        capacity_plugin = PRECHECK_PLUGINS[PrecheckPlugin.CAPACITY]()
        capacity_plugin.check()
    elif args.full_sync:
        full_sync.start(args.full_sync)
    elif args.del_config_data:
        cleanup.execute()
    elif args.command == _COMMAND_RESIZE_PVC:
        upgrade_pvc_size.resize(args)
    else:
        parser.print_help()


def in_kindtest() -> bool:
    return "KINDTEST" in os.environ


try:
    main()
except Exception as e:
    print(traceback.format_exc())
    sys.exit(1)
