#!/usr/bin/python3
"""
This script is run to upgrade the PVC size of a statefulset.

This script is intended to be idempotent, and will be run before upgrades.

:Copyright:
    Copyright 2024 VMware, Inc.  All Rights Reserved.
"""
import argparse
import datetime
import logging
import os
import re
import sys
import time
import types

from kubernetes import client, config
from kubernetes.client import rest


UNIT_DICT = types.MappingProxyType(
    {"K": 1000, "Ki": 1024, "M": 1000000, "Mi": 1048576, "G": 1000000000, "Gi": 1073741824}
)

FORMAT = "%(asctime)s %(levelname)s: %(message)s"
DATEFMT = "%Y-%m-%d %H:%M:%S"
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.INFO)


class Error(Exception):
    """Base class for exceptions in this script."""


class FailedToResizePvc(Error):
    """Failed to resize PVC"""


class FailedToParseSize(Error):
    """Failed to parse size"""


class StatefulSetNotFound(Error):
    """Statefulset was not found"""


def update_pvc_size(core_v1: client.CoreV1Api, ns: str, pvc_name: str, size: str):
    """
    Update the size of a PVC

    :param core_v1: the coreV1Api client to use
    :param ns: the namespace of the PVC
    :param pvc_name: the name of the PVC
    :param size: the size to update the PVC with

    :raises FailedToResizePvc: if resizing the PVC failed
    """
    try:
        patch = [{"op": "replace", "path": "/spec/resources/requests/storage", "value": size}]
        core_v1.patch_namespaced_persistent_volume_claim(namespace=ns, name=pvc_name, body=patch)
    except rest.ApiException as e:
        raise FailedToResizePvc(e)


def size2bytes(size: str) -> int:
    """
    Convert a string size to bytes

    :param size: the size to convert to bytes

    :return: the size in bytes
    """
    # Capture two groups, one of them has one or more numbers '\d',
    # the other has zero or more characters that are not numbers '\D'
    m = re.match(r"^(\d+)([\D]*)$", size)
    if not m:
        raise FailedToParseSize(f"Failed to parse size {size}")
    capacity = m.groups()
    number = capacity[0]
    unit = capacity[1]
    if number and unit:
        if unit not in UNIT_DICT:
            raise FailedToParseSize(f"Invalid units found for size {size}")
        size_bytes = int(number) * UNIT_DICT[unit]
    elif number:
        size_bytes = number
    else:
        raise FailedToParseSize(f"Failed to parse size {size}")

    return int(size_bytes)


def get_storage_class_allow_volume_expansion_for_sts(
    sts_name: str, ns: str, app_v1: client.AppsV1Api, storage_v1: client.StorageV1Api
) -> bool:
    """
    Check if the storage class for a statefulset allows for volume expansion

    :param sts_name: the name of the statefulset
    :param ns: the namespace of the statefulset
    :param app_v1: the AppsV1Api client to use
    :param storage_v1: the StorageV1Api client to use

    :return: True if the storageclass of the statefulset allows for volume expansion, False
        otherwise.
    """
    sts = app_v1.read_namespaced_stateful_set(namespace=ns, name=sts_name)
    storage_class_name = sts.spec.volume_claim_templates[0].spec.storage_class_name
    storage_class = storage_v1.read_storage_class(name=storage_class_name)
    if storage_class and hasattr(storage_class, "allow_volume_expansion"):
        allow_volume_expansion = storage_class.allow_volume_expansion
        return allow_volume_expansion
    else:
        logging.info(
            "StorageClass %s not found or does not support volume expansion.", storage_class_name
        )
        return False


def _pvc_ready(pvc: client.V1PersistentVolumeClaim) -> bool:
    """
    Check if the PVC is ready for the statefulset to be restarted

    :param pvc: The PVC to check

    :return: True if statefulsets using this PVC should be restarted, False otherwise.
    """
    if not pvc.status.conditions:
        return False
    for c in pvc.status.conditions:
        if c.type == "FileSystemResizePending":
            return True
    return False


def _is_pod_healthy(pod: client.V1Pod) -> bool:
    """
    Check if a pod is healthy

    :param pod: the pod to check

    :return: True if pod is healthy, False otherwise.
    """
    for c in pod.status.conditions:
        if c.type == "Ready" and c.status == "True":
            return True
    return False


def restart_statefulset(app_v1: client.AppsV1Api, name: str, ns: str):
    """
    Restart the statefulset gracefully

    This is the equivalent of doing 'kubectl rollout restart sts'

    :param app_v1: the AppsV1Api instance to use
    :param name: the name of the statefulset
    :param ns: the namespace of the statefulset
    """
    now = datetime.datetime.utcnow()
    now = str(now.isoformat("T") + "Z")
    body = {
        "spec": {
            "template": {"metadata": {"annotations": {"kubectl.kubernetes.io/restartedAt": now}}}
        }
    }
    app_v1.patch_namespaced_stateful_set(name=name, namespace=ns, body=body)


def get_statefulset(app_v1: client.AppsV1Api, name: str, ns: str):
    """
    Get a statefulset and raise if it was not found

    :param app_v1: the AppsV1Api instance to use
    :param name: the name of the statefulset
    :param ns: the namespace of the statefulset

    :raises StatefulSetNotFound: if the statefulset does not exist
    :raises ApiException: if there was an error getting the statefulset
    """
    try:
        s = app_v1.read_namespaced_stateful_set(name=name, namespace=ns)
    except rest.ApiException as e:
        if e.status == 404:
            raise StatefulSetNotFound(f"The statefulset '{name}' was not found")
        raise
    return s


def recreate_statefulset(app_v1: client.AppsV1Api, name: str, ns: str, size: str):
    """
    Recreate the statefulset to adjust the volume claim size

    :param app_v1: the AppsV1Api instance to use
    :param name: the name of the statefulset
    :param ns: the namespace of the statefulset
    :param size: create the statefulset with this volume claim size
    """
    s = get_statefulset(app_v1=app_v1, name=name, ns=ns)
    for v in s.spec.volume_claim_templates:
        v.spec.resources.requests["storage"] = size

    logging.info("Deleting statefulset")
    # Delete the statefulset but leave the pods up.
    app_v1.delete_namespaced_stateful_set(
        name=name, namespace=ns, body=client.V1DeleteOptions(propagation_policy="Orphan")
    )
    while True:
        try:
            logging.info("Waiting for statefulset '%s' to be deleted...", name)
            _ = get_statefulset(app_v1=app_v1, name=name, ns=ns)
        except StatefulSetNotFound:
            break
        time.sleep(0.5)
    # Set the resource version to None, so we can recreate the statefulset
    s.metadata.resource_version = None
    logging.info("Recreating statefulset")
    app_v1.create_namespaced_stateful_set(namespace=ns, body=s)
    # Restart the pods one by one
    logging.info("Restarting statefulset")
    restart_statefulset(app_v1=app_v1, name=name, ns=ns)


def resize_pvc_for_sts(ns: str, label: str, core_v1: client.CoreV1Api, size: str):
    """
    Resize the pvc of a statefulset

    :param sts_name: the name of the statefulset
    :param ns: the namespace of the statefulset
    :param label: the label selector to use
    :param app_v1: the AppsV1Api client to use
    :param core_v1: the CoreV1Api client to use
    :param size: the size to resize the pvc to
    """
    pvc_list = core_v1.list_namespaced_persistent_volume_claim(namespace=ns, label_selector=label)
    target_size_bytes = size2bytes(size)

    updated_pvcs = set()
    for pvc in pvc_list.items:
        try:
            current_pvc_size = pvc.spec.resources.requests["storage"]
            size_bytes = size2bytes(current_pvc_size)
        except FailedToParseSize as e:
            logging.error("Failed to parse size of pvc %s: %s", pvc.metadata.name, e)
            continue

        if size_bytes >= target_size_bytes:
            logging.info(
                "No need to resize the pvc '%s' as current size %s is GTE to %s",
                pvc.metadata.name,
                current_pvc_size,
                size,
            )
        else:
            try:
                logging.info("Resizing pvc '%s' to %s", pvc.metadata.name, size)
                update_pvc_size(core_v1, ns, pvc.metadata.name, size)
                updated_pvcs.add(pvc.metadata.name)
            except FailedToResizePvc as e:
                logging.error("Failed to resize pvc %s: %s", pvc.metadata.name, e)
                sys.exit(1)

    logging.info("Waiting for pvcs to be ready")
    any_pvcs_updated = bool(updated_pvcs)
    while any_pvcs_updated:
        pvc_list = core_v1.list_namespaced_persistent_volume_claim(
            namespace=ns, label_selector=label
        )
        if all([_pvc_ready(pvc) for pvc in pvc_list.items if pvc.metadata.name in updated_pvcs]):
            logging.info("pvcs are ready for the statefulset to be recreated")
            break
        logging.info("pvcs are not ready...")
        time.sleep(0.5)


def resize_pvc_and_recreate_statefulset(
    sts_name: str,
    ns: str,
    label: str,
    app_v1: client.AppsV1Api,
    core_v1: client.CoreV1Api,
    size: str,
):
    """
    Resize the pvc of a statefulset

    :param sts_name: the name of the statefulset
    :param ns: the namespace of the statefulset
    :param label: the label selector to use
    :param app_v1: the AppsV1Api client to use
    :param core_v1: the CoreV1Api client to use
    :param size: the size to resize the pvc to
    """
    resize_pvc_for_sts(ns=ns, label=label, core_v1=core_v1, size=size)

    target_size_bytes = size2bytes(size)
    s = get_statefulset(app_v1=app_v1, name=sts_name, ns=ns)
    try:
        storage_sizes_bytes = [
            size2bytes(v.spec.resources.requests["storage"]) for v in s.spec.volume_claim_templates
        ]
    except FailedToParseSize as e:
        logging.error("Failed to parse size of volume claim template: %s", e)
        sys.exit(1)
    need_recreate_sts = any([s < target_size_bytes for s in storage_sizes_bytes])

    if need_recreate_sts:
        logging.info("Recreating the statefulset to update the volume claim size to %s", size)
        recreate_statefulset(app_v1=app_v1, name=sts_name, ns=ns, size=size)
    else:
        logging.info(
            "Volume claims are already GTE to size %s, no need to restart statefulset", size
        )


def resize(args):
    BUILD_TAG = os.environ.get("BUILD_TAG")
    if BUILD_TAG is not None and BUILD_TAG.startswith("kindtest"):
        logging.info("Skipping pvc upgrade for kind test")
        return
    stateful_set_name = args.statefulset
    label = args.label
    size = args.size
    ns = args.namespace

    try:
        _ = size2bytes(size)
    except FailedToParseSize as e:
        sys.exit(f"Failed to parse argument 'size': {e}")

    try:
        config.load_incluster_config()
        app_v1 = client.AppsV1Api()
        core_v1 = client.CoreV1Api()
        storage_v1 = client.StorageV1Api()
        try:
            _ = get_statefulset(app_v1=app_v1, name=stateful_set_name, ns=ns)
        except StatefulSetNotFound:
            logging.info("Statefulset '%s' not found... exiting", stateful_set_name)
            sys.exit(0)

        allow_volume_expansion_enabled = get_storage_class_allow_volume_expansion_for_sts(
            stateful_set_name, ns, app_v1, storage_v1
        )

        # If storage class allow volume expansion, it will edit pvc size and restart the stateful set.
        if allow_volume_expansion_enabled:
            logging.info(
                "PVC allows for volume expansion, will resize pvc and recreate statefulset '%s'",
                stateful_set_name,
            )
            resize_pvc_and_recreate_statefulset(
                stateful_set_name, ns, label, app_v1, core_v1, size
            )
        else:
            logging.error(
                "Storage class does not allow volume expansion,"
                " unable to resize pvc for statefulset %s",
                stateful_set_name,
            )
            sys.exit(1)

    except rest.ApiException:
        logging.exception("Error upgrading pvc")
        sys.exit(1)
